# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

mixer.put() # All off.
def reset():
  for l in A_level, B_level, C_level:
    l.put(0x0d) # Half amplitude.
reset()
sleep(50) # Allow it to settle at DC 0.
for a in range(0x10):
  for b in range(0x10):
    for c in range(0x10):
      A_level.put(a)
      B_level.put(b)
      C_level.put(c)
      sleep(2)
      reset()
      sleep(2)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import AmpScale, Platform, Stream, Config
from .jackclient import BufferFiller
from .nod import Node
from .out import FloatStream, StereoInfo
from .shapes import floatdtype
from diapyr import types
from pyaudio import PyAudio, paFloat32, paContinue
import numpy as np, logging, threading

log = logging.getLogger(__name__)

class Ring: # There is a very similar ring impl in outjack, not sure if possible to unduplicate.

    def __init__(self, ringsize, newbuf, coupling):
        self.outbufs = [newbuf(np.empty) for _ in range(ringsize)]
        self.unconsumed = [False] * ringsize
        self.lock = threading.Lock()
        self.cv = threading.Condition(self.lock)
        self.readcursor = self.writecursor = 0
        self.size = ringsize
        self.coupling = coupling

    def flip(self):
        with self.lock:
            self.unconsumed[self.writecursor] = True
            self.writecursor = (self.writecursor + 1) % self.size
            logoverrun = not self.coupling
            while self.unconsumed[self.writecursor]: # Use while in case of spurious wakeup.
                if logoverrun:
                    log.error('Overrun!')
                    logoverrun = False
                self.cv.wait()
            return self.outbufs[self.writecursor]

    def consume(self, port):
        with self.lock:
            if self.unconsumed[self.readcursor]:
                np.copyto(port, self.outbufs[self.readcursor])
                self.unconsumed[self.readcursor] = False
                self.readcursor = (self.readcursor + 1) % self.size
                self.cv.notify()
            else:
                log.warning('Underrun!')

class PortAudioClient(Platform):

    @types(Config, StereoInfo)
    def __init__(self, config, stereoinfo):
        config = config.PortAudio
        self.outputrate = config.outputrate # TODO: Find best rate supported by system.
        self.buffersize = config.buffersize
        self.chancount = stereoinfo.getoutchans.size
        self.ring = Ring(config.ringsize, self._newbuf, config.coupling)
        self.port = self._newbuf(np.zeros) # Use zeros so initial underrun doesn't sound terrible.

    def _newbuf(self, constructor):
        return constructor(self.chancount * self.buffersize, dtype = floatdtype)

    def start(self):
        self.p = PyAudio()
        self.stream = self.p.open(
                rate = self.outputrate,
                channels = self.chancount,
                format = paFloat32,
                output = True,
                frames_per_buffer = self.buffersize,
                start = False,
                stream_callback = self._callback)

    def initial(self):
        return self.ring.outbufs[0]

    def flip(self):
        return self.ring.flip()

    def _callback(self, in_data, frame_count, time_info, status_flags):
        # Upstream immediately copies what we return, so assuming single thread a single port is fine:
        self.ring.consume(self.port)
        return self.port, paContinue

    def stop(self):
        self.stream.close()
        self.p.terminate()

class PortAudioStream(Node, Stream, metaclass = AmpScale):

    log2maxpeaktopeak = 1

    @types(StereoInfo, FloatStream, PortAudioClient)
    def __init__(self, stereoinfo, wavs, client):
        super().__init__()
        self.chancount = stereoinfo.getoutchans.size
        self.wavs = wavs
        self.client = client

    def start(self):
        self.filler = BufferFiller(self.chancount, self.client.buffersize, self.client.initial, self.client.flip, True)
        self.client.stream.start_stream()

    def callimpl(self):
        self.filler([self.chain(wav) for wav in self.wavs])

    def flush(self):
        pass

    def stop(self):
        self.client.stream.stop_stream()

def configure(di):
    di.add(PortAudioClient)
    di.add(PortAudioStream)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from mynblep.shapes import floatdtype
from .const import u4
import math, numpy as np

signaldtype = np.uint8 # Slightly faster than plain old int.
floatdtype = floatdtype
log2 = math.log(2)

def level5toamp(level):
    return 2 ** ((level - 31) / 4)

def amptolevel5(amp):
    return 31 + 4 * math.log(amp) / log2

def amptolevel4(amp):
    return 15 + 2 * math.log(amp) / log2

def level4to5(level4):
    return level4 * 2 + 1 # Observe 4-bit 0 is 5-bit 1.

class Shape:

    pyrbotype = dict(buf = [signaldtype], size = u4, introlen = u4)

    def __init__(self, g, introlen = 0):
        self.buf = np.fromiter(g, signaldtype)
        self.size = self.buf.size
        self.introlen = introlen

toneshape = Shape([1, 0])

def meansin(x1, x2):
    return (-math.cos(x2) - -math.cos(x1)) / (x2 - x1)

def sinsliceamp(i, n, skew):
    return (meansin(*(2 * math.pi * (i + off + skew) / n for off in [-.5, .5])) + 1) / 2

def makesinusshape(steps, maxlevel4, skew):
    maxamp = level5toamp(level4to5(maxlevel4))
    amps = [maxamp * sinsliceamp(step, steps, skew) for step in range(steps)]
    # For each step, the level that's closest to its ideal mean amp:
    unit = [max(0, int(round(amptolevel4(amp)))) for amp in amps]
    return Shape(unit)

leveltosinusshape = {level4: makesinusshape(8, level4, 0) for level4 in range(16)}



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

import logging
from .const import midichannelcount
from .iface import Config
from diapyr import types

log = logging.getLogger(__name__)

class Mediation:

    def __init__(self, config):
        self.chipchantomidipairs = tuple(set() for _ in range(config.chipchannels))
        self.midipairtochipchan = {}

    def currentmidichans(self, chipchan): # Only used for logging.
        return [midichan for midichan, _ in self.chipchantomidipairs[chipchan]]

    def acquirechipchan(self, midichan, midinote, frame):
        midipair = midichan, midinote
        if midipair in self.midipairtochipchan:
            return self.midipairtochipchan[midipair] # Spurious case.
        chipchan = self.tochipchan(midichan, frame)
        self.midipairtochipchan[midipair] = chipchan
        self.chipchantomidipairs[chipchan].add(midipair)
        return chipchan

    def releasechipchan(self, midichan, midinote):
        midipair = midichan, midinote
        if midipair in self.midipairtochipchan: # Non-spurious case.
            chipchan = self.midipairtochipchan.pop(midipair)
            self.chipchantomidipairs[chipchan].discard(midipair)
            return chipchan

    def tochipchan(self, midichan, frame):
        raise NotImplementedError

class DynamicMediation(Mediation):

    interruptingformat = "[%s] Interrupting note on channel."

    @types(Config)
    def __init__(self, config):
        super().__init__(config)
        midichanbase = config.midichannelbase
        chipchancount = config.chipchannels
        self.midichantochipchanhistory = {midichanbase + i: list(range(chipchancount)) for i in range(midichannelcount)}
        self.chipchantoonframe = [None] * chipchancount
        self.warn = log.warning

    def tochipchan(self, midichan, frame):
        def acquire():
            chipchanhistory.insert(0, chipchanhistory.pop(i))
            self.chipchantoonframe[chipchan] = frame
            return chipchan
        chipchanhistory = self.midichantochipchanhistory[midichan]
        offchipchans = {chipchan for chipchan, pairs in enumerate(self.chipchantomidipairs) if not pairs}
        if offchipchans:
            for i, chipchan in enumerate(chipchanhistory):
                if chipchan in offchipchans:
                    return acquire()
        else:
            bestonframe = min(self.chipchantoonframe) # May be None.
            bestchipchans = set(c for c, f in enumerate(self.chipchantoonframe) if f == bestonframe)
            for i, chipchan in enumerate(chipchanhistory):
                if chipchan in bestchipchans:
                    self.warn(self.interruptingformat, chr(ord('A') + chipchan))
                    for mc, mn in self.chipchantomidipairs[chipchan].copy():
                        self.releasechipchan(mc, mn)
                    return acquire()

class SimpleMediation(Mediation):

    @types(Config)
    def __init__(self, config):
        super().__init__(config)
        self.midichanbase = config.midichannelbase
        self.chipchancount = config.chipchannels

    def tochipchan(self, midichan, frame):
        return (midichan - self.midichanbase) % self.chipchancount

class PlayerMediation(Mediation):

    @types(Config)
    def __init__(self, config):
        super().__init__(config)
        self.chipchancount = config.chipchannels

    def tochipchan(self, midichan, frame):
        channel = ord(midichan[1]) - ord('1') # FIXME: Do this properly.
        return channel % self.chipchancount



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .pitch import Pitch
from .program import FX, NullNote
from .const import midichannelcount
from .mediation import Mediation
from .iface import Config, Tuning
from .ym2149 import LogicalRegisters
from diapyr import types
from contextlib import contextmanager
import logging

log = logging.getLogger(__name__)

class ChanNote:

    def __init__(self, onframe, program, nomclock, chip, chipindex, midipair, voladj, fx, normvel, tuning):
        self.program = program
        class TunedPitch(Pitch):
            def freq(self):
                return tuning.freq(self)
        with self._guard():
            self.note = program(nomclock, chip, chipindex, TunedPitch(midipair[1]), voladj, fx, normvel)
        self.onframe = onframe
        self.chip = chip
        self.chipindex = chipindex
        self.midipair = midipair
        self.fx = fx
        self.offframe = None

    def off(self, frame):
        with self._guard():
            self.note.onframes = frame - self.onframe
        self.offframe = frame

    def update(self, frame):
        with self._guard():
            self._update(frame)

    @contextmanager
    def _guard(self):
        program = self.program
        try:
            yield
        except Exception:
            log.exception("%s failed:", program.__name__)
            self.off = self.update = lambda *args: None # Freeze this note.

    def _callnoteon(self):
        if self.program.clean: # XXX: Or go via self.note?
            self.chip.flagsoff(self.chipindex) # Make it so that the impl only has to switch things on.
        self.note.noteon()

    def _update(self, frame):
        if self.offframe is None:
            f = frame - self.onframe
            if not f:
                self._callnoteon()
            self.note.noteonframe(f) # May never be called, so noteoff/noteoffframe should not rely on side-effects.
        else:
            if self.onframe == self.offframe:
                self._callnoteon()
            f = frame - self.offframe
            if not f:
                self.note.noteoff()
            self.note.noteoffframe(f)

class ControlPair:

    def __init__(self, binaryzero, flush, shift):
        self.binary = self.binaryzero = binaryzero
        self.flush = flush
        self.shift = shift

    def install(self, d, msbindex):
        d[msbindex] = self.setmsb
        d[msbindex + 0x20] = self.setlsb

    def setmsb(self, midichan, msb):
        self.binary = (msb << 7) | (self.binary & 0x7f)
        self.flush(midichan, (self.binary - self.binaryzero) >> self.shift)

    def setlsb(self, midichan, lsb):
        self.binary = (self.binary & (0x7f << 7)) | lsb
        self.flush(midichan, (self.binary - self.binaryzero) >> self.shift)

class Channels:

    @classmethod
    def configure(cls, di):
        di.add(cls)

    @types(Config, LogicalRegisters, Mediation, Tuning)
    def __init__(self, config, chip, mediation, tuning):
        self.nomclock = config.nominalclock
        neutralvel = config.neutralvelocity
        velperlevel = config.velocityperlevel
        self.tovoladj = lambda vel: (vel - neutralvel + velperlevel // 2) // velperlevel
        self.chip = chip
        self.midichantoprogram = dict(config.midichanneltoprogram) # Copy as we will be changing it.
        self.slidemidichans = set(config.slidechannels)
        self.fxfactory = lambda midichan: FX(config, midichan in self.slidemidichans)
        self.midichantofx = {c: self.fxfactory(c) for c in range(config.midichannelbase, config.midichannelbase + midichannelcount)}
        self.mediation = mediation
        self.tuning = tuning
        self.zerovelisnoteoffmidichans = set(config.zerovelocityisnoteoffchannels)
        self.monophonicmidichans = set(config.monophonicchannels)
        self.controllers = {}
        def flush(midichan, value):
            self.midichantofx[midichan].modulation.set(value)
        ControlPair(0, flush, 0).install(self.controllers, 0x01)
        def flush(midichan, value):
            self.midichantofx[midichan].pan.set(value)
        ControlPair(0x2000, flush, 0).install(self.controllers, 0x0a)
        self.prevtext = None
        self.frameindex = 0
        self.channotes = [None] * config.chipchannels
        throwawayfx = self.fxfactory(None)
        for chipchan in range(config.chipchannels):
            self.newnote(NullNote, (None, 60), 0x7f, throwawayfx, chipchan)

    @staticmethod
    def normvel(vel):
        return max(0, (vel - 1) / 0x7e)

    def newnote(self, program, midipair, vel, fx, chipchan):
        self.channotes[chipchan] = ChanNote(self.frameindex, program, self.nomclock, self.chip, chipchan, midipair, self.tovoladj(vel), fx, self.normvel(vel), self.tuning)

    def _getfx(self, midichan):
        try:
            fx = self.midichantofx[midichan]
        except KeyError:
            self.midichantofx[midichan] = fx = self.fxfactory(midichan)
        return fx

    def noteon(self, midichan, midinote, vel):
        if (not vel) and midichan in self.zerovelisnoteoffmidichans: # TODO: This is normal, confirm midi spec.
            return self.noteoff(midichan, midinote, vel)
        if midichan in self.monophonicmidichans:
            for mn in range(0x80): # FIXME LATER: We have microtonal midinotes now.
                self.noteoff(midichan, mn, 0)
        fx = self._getfx(midichan)
        if midichan in self.slidemidichans:
            fx.bend.value = 0 # Leave target and rate as-is. Note race with midi instant pitch bend (fine part 0).
        # XXX: Keep owner program for logging?
        program = self.midiprograms[self.midichantoprogram[midichan]].programformidinote(midinote) # TODO: Friendlier errors.
        chipchan = self.mediation.acquirechipchan(midichan, midinote, self.frameindex)
        self.newnote(program, (midichan, midinote), vel, fx, chipchan)
        return chipchan,

    def noteoff(self, midichan, midinote, vel): # XXX: Use vel?
        chipchan = self.mediation.releasechipchan(midichan, midinote)
        if chipchan is not None:
            channote = self.channotes[chipchan]
            if (midichan, midinote) == channote.midipair:
                channote.off(self.frameindex)
                return chipchan,

    def pitchbend(self, midichan, bend):
        self.midichantofx[midichan].bend.set(bend)

    def controlchange(self, midichan, controller, value):
        if controller in self.controllers:
            self.controllers[controller](midichan, value)

    def programchange(self, midichan, program):
        self.midichantoprogram[midichan] = program

    def updateall(self):
        text = ' | '.join("%s@%s" % (channote.program.__name__, self.mediation.currentmidichans(channote.chipindex)) for channote in self.channotes)
        if text != self.prevtext:
            log.debug(text)
            self.prevtext = text
        for channote in self.channotes:
            channote.update(self.frameindex)

    def closeframe(self):
        for fx in self.midichantofx.values():
            fx.applyrates()
        self.frameindex += 1

    def getpans(self):
        for channote in self.channotes:
            yield channote.fx.normpan()

    def __str__(self):
        return ', '.join("%s -> %s" % (midichan, self.midiprograms[program]) for midichan, program in sorted(self.midichantoprogram.items()))



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .const import clientname
from .iface import Config, Stream, Timer
from .pll import PLL
from .channels import Channels
from .util import EMA
from .speed import SpeedDetector
from .native import calsa
from diapyr import types
from diapyr.util import singleton
from splut.bg import MainBackground, SimpleBackground
import logging, time

log = logging.getLogger(__name__)

class UpdateSchedule:

    maxdelay = .01
    targetlatency = .01 # Conservative?

    def __init__(self, updaterate, skipenabled):
        self.period = 1 / updaterate
        self.taketimeema = EMA(.1, time.time())
        self.skipenabled = skipenabled

    def awaittaketime(self):
        skipped = 0
        while True:
            now = time.time()
            if now < self.taketimeema.value:
                time.sleep(self.taketimeema.value - now)
            elif self.skipenabled and now > self.taketimeema.value + self.maxdelay:
                # Forget sync with current update, try the next one:
                self.taketimeema.value += self.period
                skipped += 1
            else: # We're in the [0, maxdelay] window (or late if skip disabled).
                break
        if skipped:
            log.warning("Skipped sync with %s updates.", skipped)
        return now

    def step(self, idealtaketime):
        # TODO: Instead of EMA, improve sync with PLL.
        self.taketimeema.value += self.period
        self.taketimeema(idealtaketime + self.targetlatency)

class ChannelMessage:

    def __init__(self, config, event):
        self.midichan = config.midichannelbase + event.channel

class ChannelStateMessage(ChannelMessage): pass

class NoteOnOff(ChannelMessage):

    @staticmethod
    def formatchipchans(chipchans):
        if chipchans is not None:
            return ''.join(chr(ord('A') + chipchan) for chipchan in chipchans)

    def __init__(self, config, event):
        super().__init__(config, event)
        self.midinote = event.note
        self.vel = event.velocity

    def __str__(self):
        return "%s %2s %3s %3s" % (self.char, self.midichan, self.midinote, self.vel)

class NoteOn(NoteOnOff):

    char = 'I'

    def __call__(self, channels):
        return self.formatchipchans(channels.noteon(self.midichan, self.midinote, self.vel))

class NoteOff(NoteOnOff):

    char = 'O'

    def __call__(self, channels):
        return self.formatchipchans(channels.noteoff(self.midichan, self.midinote, self.vel))

class PitchBend(ChannelStateMessage):

    def __init__(self, config, event):
        super().__init__(config, event)
        self.bend = event.value # In [-0x2000, 0x2000).

    def __call__(self, channels):
        return channels.pitchbend(self.midichan, self.bend)

    def __str__(self):
        return "B %2d %5d" % (self.midichan, self.bend)

class ProgramChange(ChannelStateMessage):

    def __init__(self, config, event):
        super().__init__(config, event)
        self.program = config.midiprogrambase + event.value

    def __call__(self, channels):
        return channels.programchange(self.midichan, self.program)

    def __str__(self):
        return "P %2s %3s" % (self.midichan, self.program)

class ControlChange(ChannelStateMessage):

    def __init__(self, config, event):
        super().__init__(config, event)
        self.controller = event.param
        self.value = event.value

    def __call__(self, channels):
        return channels.controlchange(self.midichan, self.controller, self.value)

    def __str__(self):
        return "C %2d %3d %3d" % (self.midichan, self.controller, self.value)

class MidiListen(SimpleBackground):

    classes = {
        calsa.SND_SEQ_EVENT_NOTEON: NoteOn,
        calsa.SND_SEQ_EVENT_NOTEOFF: NoteOff,
        calsa.SND_SEQ_EVENT_PITCHBEND: PitchBend,
        calsa.SND_SEQ_EVENT_PGMCHANGE: ProgramChange,
        calsa.SND_SEQ_EVENT_CONTROLLER: ControlChange,
    }

    @types(Config, PLL)
    def __init__(self, config, pll):
        super().__init__(config.profile)
        self.pllignoremidichans = set(config.performancechannels)
        log.info("MIDI channels not significant for PLL: {%s}", ', '.join(str(c) for c in sorted(self.pllignoremidichans)))
        self.config = config
        self.pll = pll

    def start(self):
        super().start(self.bg, calsa.Client(clientname.encode(), ("%s IN" % clientname).encode()))

    def bg(self, client):
        while not self.quit:
            event = client.event_input()
            if event is not None:
                eventobj = self.classes[event.type](self.config, event)
                self.pll.event(event.time, eventobj, eventobj.midichan not in self.pllignoremidichans)

@singleton
class NullSpeedDetector:

    speedphase = None

    def __call__(self, eventcount): pass

class EventPump(MainBackground):

    @types(Config, Channels, Stream, Timer, PLL)
    def __init__(self, config, channels, stream, timer, pll):
        super().__init__(config)
        self.updaterate = config.updaterate
        self.performancemidichans = set(config.performancechannels)
        self.skipenabled = config.inputskipenabled
        self.speeddetector = SpeedDetector(10) if config.speeddetector else NullSpeedDetector
        self.channels = channels
        self.stream = stream
        self.timer = timer
        self.pll = pll

    def __call__(self):
        schedule = UpdateSchedule(self.updaterate, self.skipenabled)
        while not self.quit:
            update = self.pll.takeupdateimpl(schedule.awaittaketime())
            schedule.step(update.idealtaketime)
            scheduledevents = 0
            for event in update.events:
                if event.midichan not in self.performancemidichans:
                    scheduledevents += 1
            self.speeddetector(scheduledevents)
            timecode = self.channels.frameindex
            if self.speeddetector.speedphase is not None:
                speed = self.speeddetector.speedphase[0]
                timecode = "%s*%s+%s" % (timecode // speed, speed, timecode % speed)
            chanandnotetoevents = {}
            for event in update.events:
                if isinstance(event, NoteOnOff):
                    try:
                        chanandnotetoevents[event.midichan, event.midinote].append(event)
                    except KeyError:
                        chanandnotetoevents[event.midichan, event.midinote] = [event]
            # Apply all channel state events first:
            sortedevents = [event for event in update.events if isinstance(event, ChannelStateMessage)] # XXX: Before off?
            # Then all notes that end up off:
            for noteevents in chanandnotetoevents.values():
                if NoteOff == type(noteevents[-1]): # FIXME: May be simulated NoteOff for previous NoteOn.
                    sortedevents.extend(noteevents)
            # Then all notes that end up on:
            for noteevents in chanandnotetoevents.values():
                if NoteOn == type(noteevents[-1]):
                    sortedevents.extend(noteevents)
            # Any custom events:
            sortedevents.extend(event for event in update.events if not isinstance(event, ChannelMessage)) # XXX: Retire?
            format = "%s %.6f %s%s"
            for event in sortedevents:
                try:
                    response = event(self.channels)
                    log.debug(format, timecode, event.offset, event, '' if response is None else " -> %s" % response)
                except Exception:
                    log.exception(format, timecode, event.offset, event, '')
            self.channels.updateall()
            for block in self.timer.blocksforperiod(self.updaterate):
                self.stream.call(block)
            self.channels.closeframe()
        self.stream.flush()



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .clock import stclock
from .dac import PWMEffect, SinusEffect
from .iface import YMFile, Config
from diapyr import types
import struct, logging, os, tempfile, shutil, sys

log = logging.getLogger(__name__)

class LoopInfo:

    def __init__(self, frame, offset):
        self.frame = frame
        self.offset = offset

class YMFileException(Exception): pass

class YM:

    checkstr = b'LeOnArD!'
    wordstruct = struct.Struct('>H')
    lwordstruct = struct.Struct('>I')
    lwordlestruct = struct.Struct('<I')

    @staticmethod
    def logignoringloopinfo():
        log.debug('Ignoring loop info.')

    def __init__(self, f, expectcheckstr):
        log.debug("Format ID: %s", self.formatid)
        if expectcheckstr:
            if self.checkstr != f.read(len(self.checkstr)):
                raise YMFileException('Bad check string.')
        self.frameindex = 0
        self.f = f

    def number(self, struct):
        return struct.unpack(self.f.read(struct.size))[0]

    def word(self):
        return self.number(self.wordstruct)

    def lword(self):
        return self.number(self.lwordstruct)

    def readloopframe(self):
        loopframe = self.lword()
        if loopframe < self.framecount:
            return loopframe
        self.skip(-4)
        loopframe = self.number(self.lwordlestruct)
        if loopframe < self.framecount:
            log.info('Loop frame apparently little-endian.')
            return loopframe
        loopframe = 0
        log.warning("Cannot interpret loop frame, defaulting to %s.", loopframe)
        return loopframe

    def skip(self, n):
        self.f.seek(n, 1)

    def ntstring(self):
        start = self.f.tell()
        while ord(self.f.read(1)):
            pass
        textlen = self.f.tell() - 1 - start
        self.f.seek(start)
        text = self.f.read(textlen).decode()
        self.skip(1)
        return text

    def interleavedframe(self):
        frame = [None] * self.framesize
        for i in range(self.framesize - 1):
            frame[i] = ord(self.f.read(1))
            self.skip(-1 + self.framecount)
        frame[self.framesize - 1] = ord(self.f.read(1))
        self.skip(-(self.framesize - 1) * self.framecount)
        return frame

    def simpleframe(self):
        return [ord(c) for c in self.f.read(self.framesize)]

    def step(self):
        frame = self.frameobj(self)
        self.frameindex += 1
        return frame

    def __iter__(self):
        while self.frameindex < self.framecount:
            yield self.step()
        if self.loopinfo is None:
            return
        while True:
            if not (self.frameindex - self.framecount) % (self.framecount - self.loopinfo.frame):
                log.debug("Looping to frame %s.", self.loopinfo.frame)
                self.f.seek(self.loopinfo.offset)
            yield self.step()

    def close(self):
        self.f.close()

class PlainFrame:

    def __init__(self, ym):
        self.data = ym.readframe()

    def __call__(self, chip):
        for i, x in enumerate(self.data):
            if 0xD != i or 255 != x:
                chip.R[i].value = x

class YM23(YM):

    framesize = 14
    clock = stclock
    framefreq = 50
    info = ()
    readframe = YM.interleavedframe
    loopinfo = None # Default, overridden in YM3b.
    frameobj = PlainFrame

    def __init__(self, f):
        super().__init__(f, False)
        self.framecount = (os.fstat(f.fileno()).st_size - len(self.formatid)) // self.framesize

class YM2(YM23): # FIXME LATER: Work out format from ST-Sound source, it's not this simple.

    formatid = 'YM2!'

    def __init__(self, f, once):
        super().__init__(f)

class YM3(YM23):

    formatid = 'YM3!'

    def __init__(self, f, once):
        super().__init__(f)

class YM3b(YM23):

    formatid = 'YM3b'

    def __init__(self, f, once):
        super().__init__(f)
        if once:
            self.logignoringloopinfo()
        else:
            self.skip(self.framecount * self.framesize)
            loopframe = self.readloopframe()
            self.skip(-(self.framecount * self.framesize + 4))
            self.loopinfo = LoopInfo(loopframe, self.f.tell() + loopframe)

class YM56(YM):

    framesize = 16

    def __init__(self, f, once):
        super().__init__(f, True)
        self.framecount = self.lword()
        # We can ignore the other attributes as they are specific to sample data:
        interleaved = self.lword() & 0x01
        samplecount = self.word()
        self.clock = self.lword()
        self.framefreq = self.word()
        loopframe = self.readloopframe()
        self.skip(self.word()) # Future expansion.
        if samplecount:
            log.warning("Ignoring %s samples.", samplecount)
            for _ in range(samplecount):
                self.skip(self.lword())
        self.info = tuple(self.ntstring() for _ in range(3))
        dataoffset = self.f.tell()
        self.readframe = self.interleavedframe if interleaved else self.simpleframe
        if once:
            self.logignoringloopinfo()
            self.loopinfo = None
        elif interleaved:
            self.loopinfo = LoopInfo(loopframe, dataoffset + loopframe)
        else:
            self.loopinfo = LoopInfo(loopframe, dataoffset + loopframe * self.framesize)
        self.logdigidrum = True

class Frame56(PlainFrame):

    def __init__(self, ym):
        super().__init__(ym)
        self.index = ym.frameindex
        self.flags = ym

class Frame5(Frame56):

    def __call__(self, chip):
        super().__call__(chip)
        if self.data[0x1] & 0x30:
            chan = ((self.data[0x1] & 0x30) >> 4) - 1
            tcr = (self.data[0x6] & 0xe0) >> 5
            if tcr:
                tdr = self.data[0xE]
                chip.timers[chan].update(tcr, tdr, PWMEffect(chip.R[chip.levelbase + chan]))
        if self.flags.logdigidrum and (self.data[0x3] & 0x30):
            log.warning("Digi-drum at frame %s.", self.index)
            self.flags.logdigidrum = False

class Frame6(Frame56):

    def __call__(self, chip):
        super().__call__(chip)
        timerchans = set()
        for r, rr, rrr in [0x1, 0x6, 0xE], [0x3, 0x8, 0xF]:
            if self.data[r] & 0x30:
                chan = ((self.data[r] & 0x30) >> 4) - 1
                fx = self.data[r] & 0xc0
                if 0x00 == fx:
                    tcr = (self.data[rr] & 0xe0) >> 5
                    if tcr:
                        tdr = self.data[rrr]
                        chip.timers[chan].update(tcr, tdr, PWMEffect(chip.R[chip.levelbase + chan]))
                        timerchans.add(chan)
                if self.flags.logdigidrum and 0x40 == fx:
                    log.warning("Digi-drum at frame %s.", self.index)
                    self.flags.logdigidrum = False
                if 0x80 == fx:
                    tcr = (self.data[rr] & 0xe0) >> 5
                    if tcr:
                        tdr = self.data[rrr]
                        chip.timers[chan].update(tcr, tdr, SinusEffect(chip.R[chip.levelbase + chan]))
                        timerchans.add(chan)
                if self.flags.logsyncbuzzer and 0xc0 == fx:
                    log.warning("Sync-buzzer at frame %s.", self.index)
                    self.flags.logsyncbuzzer = False
        for chan, timer in enumerate(chip.timers):
            if chan not in timerchans:
                timer.effect.value = None

class YM5(YM56):

    formatid = 'YM5!'
    frameobj = Frame5

class YM6(YM56):

    formatid = 'YM6!'
    frameobj = Frame6

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.logsyncbuzzer = True

class YMOpen(YMFile):

    impls = {i.formatid.encode(): i for i in [YM2, YM3, YM3b, YM5, YM6]}

    @types(Config)
    def __init__(self, config):
        self.path = config.inpath
        self.once = config.ignoreloop

    def start(self):
        self.startimpl()
        for info in self.ym.info:
            log.info(info)
        self.nominalclock = self.ym.clock
        self.pianorollheight = self.updaterate = self.ym.framefreq

    def startimpl(self):
        self.f = open(self.path, 'rb')
        try:
            if 'YM' == self.f.read(2):
                self.ym = self.impls['YM' + self.f.read(2)](self.f, self.once)
                return
        except:
            self.f.close()
            raise
        self.f.close()
        self.f = UnpackedFile(self.path)
        try:
            self.ym = self.impls[self.f.read(4)](self.f, self.once)
            # return
        except:
            self.f.close()
            raise

    def frames(self, chip):
        for frame in self.ym:
            frame(chip)
            yield

    def stop(self):
        self.f.close()

class UnpackedFile:

    def __init__(self, path):
        self.tmpdir = tempfile.mkdtemp()
        try:
            from lagoon import lha
            # Observe we redirect stdout so it doesn't get played:
            lha.x(os.path.abspath(path), cwd = self.tmpdir, stdout = sys.stderr)
            name, = os.listdir(self.tmpdir)
            self.f = open(os.path.join(self.tmpdir, name), 'rb')
        except:
            self.clean()
            raise

    def clean(self):
        log.debug("Deleting temporary folder: %s", self.tmpdir)
        shutil.rmtree(self.tmpdir)

    def __getattr__(self, name):
        return getattr(self.f, name)

    def close(self):
        self.f.close()
        self.clean()



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config
from argparse import ArgumentParser
from aridity import NoSuchPathException
from aridity.model import Number, Text
from diapyr import UnsatisfiableRequestException
from importlib import import_module
from pathlib import Path
import aridity.config, logging, lurlene, numbers, sys

log = logging.getLogger(__name__)
namespace = 'pym2149'

class ConfigName:

    def __init__(self, *params, args = sys.argv[1:], name = 'defaultconf'):
        parser = ArgumentParser()
        parser.add_argument('--repr', action = 'append', default = [])
        parser.add_argument('--config', action = 'append', default = [])
        parser.add_argument('--ignore-settings', action = 'store_true')
        for param in params:
            parser.add_argument(param)
        self.additems = parser.parse_args(args)
        self.path = Path(__file__).resolve().parent / ("%s.arid" % name)

    def applyitems(self, config):
        if not self.additems.ignore_settings:
            try:
                config.loadsettings()
            except FileNotFoundError as e:
                log.warning("Settings not found: %s", e)
        for name, value in self.additems.__dict__.items():
            if 'config' == name:
                with config.repl() as repl:
                    for text in value:
                        repl.printf("%s", namespace)
                        for line in text.splitlines():
                            repl("\t%s" % line)
            else:
                config.put(namespace, name, resolvable = wrap(value))

    def newloader(self, di):
        return ConfigLoader(self, di)

def wrap(value): # TODO: Migrate to aridity.
    return (Number if isinstance(value, numbers.Number) else Text)(value)

class AsContext:

    def __init__(self, parent, obj):
        self.parent = parent
        self.obj = obj

    def resolved(self, name):
        try:
            return wrap(getattr(self.obj, name))
        except AttributeError:
            return self.parent.resolved(name)

class ConfigLoader:

    @staticmethod
    def getglobal(context, resolvable):
        spec = resolvable.resolve(context).cat()
        lastdot = spec.rindex('.')
        return wrap(getattr(import_module(spec[:lastdot], __package__), spec[lastdot + 1:]))

    @staticmethod
    def enter(context, contextresolvable, resolvable):
        return resolvable.resolve(contextresolvable.resolve(context))

    @staticmethod
    def py(config, context, *clauses):
        return wrap(eval(' '.join(c.cat() for c in clauses), dict(config = config)))

    def resolve(self, context, resolvable):
        try:
            return AsContext(context, self.di(self.getglobal(context, resolvable).value))
        except UnsatisfiableRequestException:
            raise NoSuchPathException

    def __init__(self, configname, di):
        self.configname = configname
        self.di = di

    def mark(self):
        path = self.configname.path
        self.mtime = path.stat().st_mtime
        return path

    def load(self):
        config = ConfigImpl.blank()
        config.put('global', function = self.getglobal)
        config.put('enter', function = self.enter)
        config.put('py', function = lambda *args: self.py(config, *args))
        config.put('resolve', function = self.resolve)
        with config.repl() as repl:
            path = self.mark()
            repl.printf("cwd = %s", path.parent)
            repl.printf("%s . %s", namespace, path.name)
        self.configname.applyitems(config)
        config = getattr(config, namespace)
        return config

class ConfigImpl(aridity.config.Config, Config, lurlene.iface.Config): pass



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .lfo import LFO, FloatLFO
from .reg import Reg, regproperty
import logging

log = logging.getLogger(__name__)

class FX:

    class SimpleValue:

        def __init__(self, value):
            self.value = value

        def set(self, value):
            self.value = value

        def step(self): pass

    class SlideValue:

        def __init__(self, slidespeed, value):
            self.slidespeed = slidespeed
            self.target = self.value = value

        def set(self, value):
            self.target = value & ~0x7f
            self.rate = (value & 0x7f) * self.slidespeed
            if not self.rate: # Simulate simple behaviour when fine part is 0.
                self.value = self.target

        def step(self):
            if self.value < self.target:
                self.value = min(self.value + self.rate, self.target)
            elif self.value > self.target:
                self.value = max(self.value - self.rate, self.target)

    halfrange = 1 << 13 # Full range is 14 bits.
    minsigned = -halfrange
    maxsigned = halfrange - 1

    @staticmethod
    def signum(n):
        return 1 if n > 0 else (-1 if n < 0 else 0)

    def __init__(self, config, slide):
        self.bendpersemitone = config.pitchbendpersemitone
        if slide:
            slidespeed = config.slidespeed
            valueimpl = lambda v: self.SlideValue(slidespeed, v)
        else:
            valueimpl = self.SimpleValue
        self.bend = valueimpl(0)
        self.modulation = valueimpl(self.halfrange)
        self.pan = valueimpl(0)

    def applyrates(self):
        for v in self.bend, self.modulation, self.pan:
            v.step()

    def bendsemitones(self):
        return self.bend.value / self.bendpersemitone

    def relmodulation(self):
        return (max(1, self.modulation.value) - self.halfrange) / (self.halfrange - 1) / 2 + .5

    def normpan(self):
        # Observe we don't apply maxpan, which is only for the auto-stereo:
        return max(self.minsigned + 1, self.pan.value) / (self.halfrange - 1)

class Unpitched:

    @classmethod
    def programformidinote(cls, midinote):
        return cls.midinotetoprogram.get(midinote, Note) # Observe a miss results in silence.

class Note(Unpitched):

    clean = True
    toneperiod = regproperty(lambda note: note.toneperiodreg)
    toneflag = regproperty(lambda note: note.toneflagreg)
    noiseflag = regproperty(lambda note: note.noiseflagreg)
    fixedlevel = regproperty(lambda note: note.fixedlevelreg)
    levelmode = regproperty(lambda note: note.levelmodereg)
    tonepitch = regproperty(lambda note: note.tonepitchreg)
    tonefreq = regproperty(lambda note: note.tonefreqreg)

    @classmethod
    def programformidinote(cls, midinote):
        return cls

    def __init__(self, nomclock, chip, chipchan, pitch, voladj, fx, normvel):
        self.toneperiodreg = chip.toneperiods[chipchan]
        self.toneflagreg = chip.toneflags[chipchan]
        self.timer = chip.timers[chipchan]
        self.noiseflagreg = chip.noiseflags[chipchan]
        self.fixedlevelreg = Reg()
        # No reverse link, we don't want to pollute the chip with references:
        chip.fixedlevels[chipchan].link(lambda unclamped: max(0, min(15, unclamped)), self.fixedlevelreg)
        self.levelmodereg = chip.levelmodes[chipchan]
        self.tonefreqreg = Reg()
        self.toneperiodreg.link(lambda f: f.toneperiod(nomclock), self.tonefreqreg) # No reverse link.
        self.tonepitchreg = Reg()
        self.tonefreqreg.link(lambda p: (p + fx.bendsemitones()).freq(), self.tonepitchreg)
        self.nomclock = nomclock
        self.chip = chip
        self.chipchan = chipchan
        self.pitch = pitch
        self.voladj = voladj
        self.fx = fx
        self.velocity = normvel

    def noteon(self): pass

    def noteonframe(self, frame):
        "May never be called, so don't make changes that noteoff or a custom impl of noteoffframe later relies on."

    def noteoff(self): pass

    def noteoffframe(self, frame):
        self.noteonframe(self.onframes + frame)

class NullNote(Note):

    clean = False

class DefaultNote(Note): # TODO: Does not belong here.

    vib = FloatLFO(0).hold(10).tri(8, 2, .2).loop(8)
    fadeout = LFO(-1).hold(1).lin(28, -15)

    def noteon(self):
        self.toneflag = True
        self.fixedlevel = 13 + self.voladj

    def noteonframe(self, frame):
        self.tonepitch = self.pitch + self.fx.relmodulation() * self.vib(frame)

    def noteoffframe(self, frame):
        self.tonepitch = self.pitch + self.fx.relmodulation() * self.vib(self.onframes + frame)
        self.fixedlevel = 13 + self.voladj + self.fadeout(frame)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from . import E, V
from pym2149.dac import SinusEffect
from spectrogram import silence

# TODO LATER: Find out why the result appears shifted a few samples to the right.
class Sin1k:

    level = V('75x15//,0')

    def on(self, ym, frame):
        ym.level = self.level[frame]
        if frame < 1:
            chan = ym[0]._chan
            timer = ym._chip.timers[chan]
            timer.effect.value = SinusEffect(ym._chip.fixedlevels[chan])
            timer.freq.value = 1000 * 4 # FIXME: It should know wavelength from effect.

sections = [[E(Sin1k, '2'), silence, silence]]
speed = 50



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .iface import Config
from diapyr import types
from .util import EMA
import time, logging

log = logging.getLogger(__name__)

class Update:

    def __init__(self, events, idealtaketime):
        self.events = events
        self.idealtaketime = idealtaketime

class PLL:

    @types(Config)
    def __init__(self, config):
        self.updateperiod = 1 / config.updaterate
        self.targetpos = float(config.plltargetpos)
        self.medianshiftema = EMA(float(config.pllalpha), None)

    def start(self):
        self.events = []
        self.updates = []
        self.medianshiftema.value = 0
        self.mark = time.time()
        self.windowindex = 0
        self.nextwindow()

    def stop(self):
        pass

    def nextwindow(self):
        self.windowindex += 1
        self.exclusivewindowend = self.mark + self.windowindex * self.updateperiod + self.medianshiftema.value

    def event(self, eventtime, event, significant):
        self.events.append((eventtime, event, significant))

    def closeupdate(self):
        inclusivewindowstart = self.exclusivewindowend - self.updateperiod
        targettime = inclusivewindowstart + self.targetpos
        shifts = []
        i = 0
        for eventtime, _, significant in self.events:
            if eventtime >= self.exclusivewindowend:
                break
            if significant and eventtime >= inclusivewindowstart:
                shifts.append(self.medianshiftema.value + eventtime - targettime)
            # If eventtime < inclusivewindowstart we consume the event without harvesting its shift.
            i += 1
        update = []
        for eventtime, event, _ in self.events[:i]:
            event.offset = eventtime - inclusivewindowstart
            update.append(event)
        self.updates.append(update)
        del self.events[:i]
        if shifts:
            n = len(shifts)
            if n & 1: # Odd.
                medianshift = shifts[(n - 1) // 2]
            else:
                midindex = n // 2
                medianshift = (shifts[midindex - 1] + shifts[midindex]) / 2
            self.medianshiftema(medianshift)
        self.nextwindow()

    def takeupdate(self):
        return self.takeupdateimpl(time.time())

    def takeupdateimpl(self, now):
        while now >= self.exclusivewindowend: # No more events can qualify for this window.
            self.closeupdate()
        copy = self.updates[:]
        n = len(copy)
        del self.updates[:n] # Not actually necessary to use n, as only we call closeupdate.
        if 1 == n:
            update, = copy
        else:
            log.warning("Yielding %s updates as one.", n)
            update = sum(copy, [])
        return Update(update, self.exclusivewindowend)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .nod import BufNode
from pyrbo import turbo, LOCAL
from .const import u1, u4, i4, u8, i8
from .shapes import Shape, toneshape, signaldtype
import numpy as np, itertools

oscnodepyrbotype = dict(
    blockbuf = dict(buf = [signaldtype]),
    block = dict(framecount = u4),
    index = i4,
    shape = Shape.pyrbotype,
)

class ShapeOsc(BufNode):

    progressdtype = u4

    def __init__(self, scale, periodreg):
        super().__init__(signaldtype)
        self.stepsize = 0 # XXX: Move to reset?
        self.scale = scale
        self.periodreg = periodreg

    def reset(self, shape):
        self.index = -1
        self.progress = np.iinfo(self.progressdtype).max
        self.shape = shape

    def callimpl(self):
        self.index, self.progress, self.stepsize = self.shapeimpl()

    @turbo(
        self = dict(
            oscnodepyrbotype,
            progress = progressdtype,
            stepsize = u4,
            scale = u4,
            periodreg = dict(value = u4),
            eager = u1,
        ),
        i = u4,
        j = u4,
        n = u4,
        val = signaldtype,
    )
    def shapeimpl(self):
        self_blockbuf_buf = self_block_framecount = self_index = self_progress = self_scale = self_periodreg_value = self_shape_buf = self_shape_size = self_shape_introlen = self_eager = LOCAL
        if self_eager:
            self_stepsize = self_periodreg_value * self_scale
        i = 0
        if self_progress < self_stepsize:
            val = self_shape_buf[self_index]
            j = min(self_stepsize - self_progress, self_block_framecount)
            while i < j:
                self_blockbuf_buf[i] = val
                i += 1
        if i == self_block_framecount:
            self_progress += self_block_framecount
        else:
            if not self_eager:
                self_stepsize = self_periodreg_value * self_scale
            n = (self_block_framecount - i) // self_stepsize
            while n:
                self_index += 1
                if self_index == self_shape_size:
                    self_index = self_shape_introlen
                val = self_shape_buf[self_index]
                j = i + self_stepsize
                while i < j:
                    self_blockbuf_buf[i] = val
                    i += 1
                n -= 1
            if i == self_block_framecount:
                self_progress = self_stepsize
            else:
                self_index += 1
                if self_index == self_shape_size:
                    self_index = self_shape_introlen
                val = self_shape_buf[self_index]
                self_progress = self_block_framecount - i
                while i < self_block_framecount:
                    self_blockbuf_buf[i] = val
                    i += 1
        return self_index, self_progress, self_stepsize

class IncompatibleShapeException(Exception): pass

class RToneOsc(BufNode):

    def __init__(self, mfpclock, chipimplclock, timer):
        super().__init__(signaldtype)
        self.effectversion = None
        self.mfpclock = mfpclock
        self.chipimplclock = chipimplclock
        self.timer = timer

    def reset(self, shape):
        self.index = -1
        self.maincounter = 0
        self.precounterxmfp = None
        self.shape = shape

    def callimpl(self):
        if self.effectversion != self.timer.effect.version:
            self.reset(self.timer.effect.value.getshape())
            self.effectversion = self.timer.effect.version
        else:
            shape = self.timer.effect.value.getshape()
            if shape.size != self.shape.size or shape.introlen != self.shape.introlen:
                raise IncompatibleShapeException
            self.shape = shape
        prescalerornone = self.timer.prescalerornone.value
        if prescalerornone is None:
            self.blockbuf.fill_same(self.shape.buf[self.index])
            self.precounterxmfp = None
        else:
            if self.precounterxmfp is None:
                self.precounterxmfp = prescalerornone * self.chipimplclock
            self.index, self.maincounter, self.precounterxmfp = self.rtoneimpl(prescalerornone, self.timer.effectivedata.value)

    @turbo(
        self = dict(
            oscnodepyrbotype,
            mfpclock = i8, # Signed so j doesn't blow up.
            chipimplclock = u8,
            maincounter = i4,
            precounterxmfp = u4,
        ),
        prescaler = u4,
        etdr = u4,
        chunksizexmfp = u8,
        stepsizexmfp = u8,
        nextstepxmfp = i8,
        i = u4,
        j = u4,
        val = signaldtype,
    )
    def rtoneimpl(self, prescaler, etdr):
        self_blockbuf_buf = self_block_framecount = self_mfpclock = self_chipimplclock = self_index = self_maincounter = self_precounterxmfp = self_shape_buf = self_shape_size = self_shape_introlen = LOCAL
        chunksizexmfp = self_chipimplclock * prescaler
        stepsizexmfp = chunksizexmfp * etdr
        nextstepxmfp = chunksizexmfp * self_maincounter + self_precounterxmfp - chunksizexmfp
        i = 0
        while True:
            j = min((nextstepxmfp + self_mfpclock - 1) // self_mfpclock, self_block_framecount)
            val = self_shape_buf[self_index]
            while i < j:
                self_blockbuf_buf[i] = val
                i += 1
            if j == self_block_framecount:
                break
            nextstepxmfp += stepsizexmfp
            self_index += 1
            if self_index == self_shape_size:
                self_index = self_shape_introlen
        nextstepxmfp -= self_mfpclock * self_block_framecount
        self_maincounter = 1
        while nextstepxmfp < 0:
            nextstepxmfp += chunksizexmfp
            self_maincounter -= 1
        self_maincounter += nextstepxmfp // chunksizexmfp
        return self_index, self_maincounter, nextstepxmfp % chunksizexmfp

class ToneOsc(ShapeOsc):

    eager = True

    def __init__(self, scale, periodreg):
        scaleofstep = scale * 2 // 2 # Normally half of 16.
        super().__init__(scaleofstep, periodreg)
        self.reset(toneshape)

class NoiseOsc(ShapeOsc):

    eager = False

    def __init__(self, scale, periodreg, shape):
        scaleofstep = scale * 2 # This results in authentic spectrum, see qnoispec.
        super().__init__(scaleofstep, periodreg)
        self.reset(shape)

class EnvOsc(ShapeOsc):

    eager = True
    steps = 32
    shapes = {
        0x0c: Shape(range(steps)),
        0x08: Shape(range(steps - 1, -1, -1)),
        0x0e: Shape(itertools.chain(range(steps), range(steps - 1, -1, -1))),
        0x0a: Shape(itertools.chain(range(steps - 1, -1, -1), range(steps))),
        0x0f: Shape(itertools.chain(range(steps), steps * [0]), steps),
        0x0d: Shape(itertools.chain(range(steps), steps * [steps - 1]), steps),
        0x0b: Shape(itertools.chain(range(steps - 1, -1, -1), steps * [steps - 1]), steps),
        0x09: Shape(itertools.chain(range(steps - 1, -1, -1), steps * [0]), steps),
    }
    for s in range(0x08):
        shapes[s] = shapes[0x0f if s & 0x04 else 0x09]
    del s

    def __init__(self, scale, periodreg, shapereg):
        scaleofstep = scale * 32 // self.steps
        super().__init__(scaleofstep, periodreg)
        self.shapeversion = None
        self.shapereg = shapereg

    def callimpl(self):
        if self.shapeversion != self.shapereg.version:
            self.reset(self.shapes[self.shapereg.value])
            self.shapeversion = self.shapereg.version
        super().callimpl()



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from decimal import Decimal, ROUND_HALF_DOWN, ROUND_HALF_UP

class AbstractLFO:

    def __init__(self, initial):
        self.v = [initial]
        self.looplenornone = 1

    def lin(self, n, target):
        source = self.v[-1]
        for i in range(1, n + 1):
            self.v.append(source + (target - source) * i / n)
        return self

    def jump(self, target):
        self.v[-1] = target
        return self

    def then(self, *vals):
        self.v.extend(vals)
        return self

    def hold(self, n):
        for _ in range(n):
            self.v.append(self.v[-1])
        return self

    def tri(self, trin, linn, target):
        if 0 != trin % 4:
            raise Exception("Expected a multiple of 4 but got: %s" % trin)
        normn = trin // 4
        if 0 != normn % linn:
            raise Exception("Expected a factor of %s but got: %s" % (normn, linn))
        source = self.v[-1]
        for _ in range(normn // linn):
            self.lin(linn, target)
            self.lin(linn * 2, source * 2 - target)
            self.lin(linn, source)
        return self

    def loop(self, n = None):
        self.looplenornone = n
        return self

    def get(self, frame):
        n = len(self.v)
        if frame >= n:
            looplen = n if self.looplenornone is None else self.looplenornone
            start = n - looplen
            frame = start + ((frame - start) % looplen)
        return self.v[frame]

    def render(self, n = None):
        if n is None:
            n = len(self.v)
        return [self(i) for i in range(n)] # Observe via the xform.

class LFO(AbstractLFO):

    def __call__(self, frame):
        current, next = self.get(frame), self.get(frame + 1)
        if current < 0:
            towards0 = (current < next) if (next < 0) else True
        else:
            towards0 = True if (next < 0) else (next < current)
        rounding = ROUND_HALF_DOWN if towards0 else ROUND_HALF_UP
        return int(Decimal(current).to_integral_value(rounding))

class FloatLFO(AbstractLFO):

    def __call__(self, frame):
        return self.get(frame)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

import logging, numpy as np
from collections import namedtuple
from pyrbo import turbo
from .const import u4

log = logging.getLogger(__name__)

def defaultcallback(oldspeedphaseornone, speedphase, clarity):
    if oldspeedphaseornone is None:
        log.info("Speed detected: %s", speedphase)
    else:
        log.warning("Speed was %s but is now: %s", oldspeedphaseornone, speedphase)

kernelperiods = 2
dtype = np.int32 # Must be signed.
minclarity = 1.1

ScoreSpeedPhase = namedtuple('ScoreSpeedPhase', 'score speed phase')

@turbo(kernel = [dtype], kernelsize = u4, history = [dtype], maxphase = u4, s = dtype, score = dtype, i = u4, j = u4, p = u4, phase = u4)
def correlate(kernel, kernelsize, history, maxphase):
    phase = 0
    score = 0
    j = maxphase
    for i in range(kernelsize):
        score += kernel[i] * history[j]
        j += 1
    for p in range(1, maxphase + 1):
        s = 0
        j = maxphase - p
        for i in range(kernelsize):
            s += kernel[i] * history[j]
            j += 1
        if s > score:
            phase = p
            score = s
    return phase, score

class Shape:

    def __init__(self, speed):
        self.kernelsize = kernelperiods * speed + 1
        def g():
            for i in range(self.kernelsize):
                yield -1 if i % speed else 1
        self.kernel = np.fromiter(g(), dtype = dtype)
        self.speed = speed

    def bestscorephase(self, history, index):
        phase, score = correlate(self.kernel, self.kernelsize, history, self.speed - 1)
        return ScoreSpeedPhase(score, self.speed, (index + phase + 1) % self.speed)

class SpeedDetector:

    def __init__(self, maxspeed, callback = defaultcallback):
        self.shapes = [Shape(speed) for speed in range(1, maxspeed + 1)]
        self.history = np.zeros(maxspeed * (kernelperiods + 1), dtype = dtype)
        self.speedphase = None
        self.index = 0
        self.callback = callback

    def __call__(self, eventcount):
        self.history[1:] = self.history[:-1] # Lose the oldest value.
        self.history[0] = eventcount # New newest value.
        scorespeedphasetuples = sorted(shape.bestscorephase(self.history, self.index) for shape in self.shapes)
        self.index += 1
        best = scorespeedphasetuples[-1]
        if scorespeedphasetuples[-2].score:
            clarity = scorespeedphasetuples[-1].score / scorespeedphasetuples[-2].score
            if clarity >= minclarity and (best.speed, best.phase) != self.speedphase:
                oldspeedphase = self.speedphase
                self.speedphase = best.speed, best.phase
                self.callback(oldspeedphase, self.speedphase, clarity)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from pym2149.initlogging import logging
from pym2149 import out
from pym2149.boot import boot
from pym2149.config import ConfigName
from pym2149.timerimpl import ChipTimer
from pym2149.util import MainThread
from pym2149.ymformat import YMOpen
from pym2149.ymplayer import Player, PhysicalBundle
from diapyr.start import Started

log = logging.getLogger(__name__)

def main_ym2wav():
    config, di = boot(ConfigName('inpath', 'outpath'))
    with di:
        di.add(YMOpen)
        out.configure(di)
        di.add(ChipTimer)
        di.add(PhysicalBundle)
        di.add(Player)
        di.all(Started)
        di(MainThread).sleep()



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

from .nod import BufNode
from .shapes import toneshape, leveltosinusshape, level5toamp, level4to5, signaldtype, floatdtype
import numpy as np

class Level(BufNode):

    pwmzero4bit = 0 # TODO: Currently consistent with ST-Sound, but make it a register.
    pwmzero5bit = level4to5(pwmzero4bit)
    lookup = np.fromiter([pwmzero5bit] + list(range(32)), signaldtype)

    def __init__(self, levelmodereg, fixedreg, env, signal, rtone, timereffectreg):
        super().__init__(signaldtype) # Must be suitable for use as index downstream.
        self.levelmodereg = levelmodereg
        self.fixedreg = fixedreg
        self.env = env
        self.signal = signal
        self.rtone = rtone
        self.timereffectreg = timereffectreg

    def callimpl(self):
        levelmode = self.levelmodereg.value
        timereffect = self.timereffectreg.value
        if timereffect is not None:
            timereffect(levelmode, self.env, self.signal, self.rtone, self.blockbuf, self.chain)
        elif levelmode:
            self.blockbuf.copybuf(self.chain(self.signal))
            self.blockbuf.mulbuf(self.chain(self.env))
        else:
            self.blockbuf.copybuf(self.chain(self.signal))
            self.blockbuf.mul(level4to5(self.fixedreg.value))

class TimerEffect:

    def __init__(self, fixedreg):
        self.fixedreg = fixedreg

class PWMEffect(TimerEffect):

    def getshape(self):
        return toneshape

    def __call__(self, levelmode, envnode, signalnode, rtonenode, blockbuf, chain):
        if levelmode:
            # TODO: Test this branch, or override levelmode starting with first interrupt.
            blockbuf.copybuf(chain(envnode)) # Values in [0, 31].
            blockbuf.add(1) # Shift env values to [1, 32].
            blockbuf.mulbuf(chain(signalnode)) # Introduce 0.
            blockbuf.mulbuf(chain(rtonenode)) # Introduce more 0.
            blockbuf.mapbuf(blockbuf, Level.lookup) # Map 0 to 5-bit pwmzero and sub 1 from rest.
        else:
            blockbuf.copybuf(chain(signalnode))
            blockbuf.mulbuf(chain(rtonenode))
            # Map 0 to pwmzero and 1 to fixed level:
            blockbuf.mul(level4to5(self.fixedreg.value) - Level.pwmzero5bit)
            blockbuf.add(Level.pwmzero5bit)

class SinusEffect(TimerEffect):

    def getshape(self):
        return leveltosinusshape[self.fixedreg.value]

    def __call__(self, levelmode, envnode, signalnode, rtonenode, blockbuf, chain):
        blockbuf.copybuf(chain(rtonenode))
        blockbuf.mul(2)
        blockbuf.add(1)

class Dac(BufNode):

    def __init__(self, level, log2maxpeaktopeak, ampshare):
        super().__init__(floatdtype)
        # We take off .5 so that the peak amplitude is about -3 dB:
        maxpeaktopeak = (2 ** (log2maxpeaktopeak - .5)) / ampshare
        # Lookup of ideal amplitudes:
        self.leveltopeaktopeak = np.fromiter((level5toamp(v) * maxpeaktopeak for v in range(32)), self.dtype)
        self.level = level

    def callimpl(self):
        self.blockbuf.mapbuf(self.chain(self.level), self.leveltopeaktopeak)



# Copyright 2014, 2018, 2019 Andrzej Cichocki

# This file is part of pym2149.
#
# pym2149 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# pym2149 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pym2149.  If not, see <http://www.gnu.org/licenses/>.

cdef extern from "sys/time.h":

    ctypedef long time_t

    ctypedef long suseconds_t

    cdef struct timeval:
        time_t tv_sec
        suseconds_t tv_usec

    cdef struct timezone:
        pass # Obsolete.

    int gettimeofday(timeval*, timezone*) nogil
